//DL2KHP GPSDO
//Einstellen eines NEO GPS Receivers auf unterschiedliche Timepulsefrequenzen in Z.30 und Z.33
//Ausgabe der Anzahl an empfangenen Satelliten und welche Art von GPS fix vorliegt
//Einstellen des GPS RX auf stationären Betrieb und deaktivieren des SBAS Features zur Erhöhung der Timepulsegenauigkeit
//Abspeichern dieser betriebsparameter im GPS RX (falls dort ein eeprom vorhanden ist).
//Ausgabe der EFC Regelspannung vom OCXO; Messung an A0 mit oversampling auf 15 Bit ==> Auflösung der Referenzspannung von 5V auf 0,15mV im LSB
//Anzeige auf drei Nachkommastellen (1mV)
//Ausgaben an LCD und Seriell nur bei Änderungen der Parameter, damit das LCD Display nich flackert
//Display: I2C LCD mit 2 Zeilen zu je 16 zeichen




#include <TinyGPS++.h>
#include <SoftwareSerial.h>
#include <LiquidCrystal_I2C.h>


/*
   Neo M8N gps Empfänger
   Frequenz am 1pps Ausgang einstellen
   1Hz bis ca. 12 MHz
   Anzeige des Fixes (no fix, 2d fix, 3d fix) über LEDs und serieller Ausgabe sowie der Zahl
   an empfanener Satelliten über serielle Ausgabe
*/
//Globale Variablen und Konstanten

//Frequenz einstellen
//Frequenz des 1pps ohne gps fix in Herz
unsigned long frequency_no_fix = 100000;

//Frequenz des 1pps mit gps fix in Herz
unsigned long frequency_fix = 100000;

byte bytes_no_fix[4];
byte bytes_fix[4];

//Softseriell einstellen und Baudrate von sfotseriell und hardware seriell
static const int RXPin = 8, TXPin = 9;          //pins for software serial
static const uint32_t GPSBaud = 9600;       //baud rate for software serial
static const uint32_t SerialBaud = 9600;  //baud rate for hardware serial

//pins für leds
int no_fix_led = 10, two_d_fix_led = 11, three_d_fix_led = 12;    //output pins for the leds

//allg. Variablen
char fix2d3d = '0';
char fix2d3d_old = '0';
int sats;
int sats_old;
float ocxo_spannung_old  = -999.0;

// set the LCD address to 0x27 for a 16 chars 2 line display
LiquidCrystal_I2C lcd(0x27, 16, 2);

// The TinyGPS++ object
TinyGPSPlus gps;
void getgps(TinyGPSPlus &gps);

// The serial connection to the GPS device
SoftwareSerial ss(RXPin, TXPin);

//auswertung des fix zustandes (custom sequence)
TinyGPSCustom fix(gps, "GPGSA", 2); // $GPGSA sentence, 2th element: char '1' = no fix, char '2' = 2d fix, char '3' = 3d fix

void setup()
{
  Serial.begin(SerialBaud);
  ss.begin(GPSBaud);

  pinMode(no_fix_led, OUTPUT);          // sets the digital pin as output
  pinMode(two_d_fix_led, OUTPUT);          // sets the digital pin as output
  pinMode(three_d_fix_led, OUTPUT);          // sets the digital pin as output
  digitalWrite(no_fix_led, LOW);
  digitalWrite(two_d_fix_led, LOW);
  digitalWrite(three_d_fix_led, LOW);

  // initialize the LCD
  lcd.begin();

  lcd.backlight();

  lcd.setCursor(0, 0);         // move cursor to   (0, 0) 1. Zeichen, Zeile 1
  lcd.print("     GPSDO");        // print message at (0, 0)
  lcd.setCursor(0, 1);         // move cursor to   (0, 1)1. Zeichen, Zeile 1
  lcd.print("setze Timepulse "); // print message at (2, 1)

  Serial.println("GPS RX wird initialisiert....einen Moment");
  Serial.println("Frequenz 1pps wird eingestellt");
  delay(2000);

  //array mit den daten zur einstellung der frequenzen am 1pps ausgang ohne abschließende checksumme (2 bytes, ckA und ckB))
  //Aufteilen der Frequenzworte in vier Bytes zum übertragen an den neo gps rx
  bytes_no_fix[0] = frequency_no_fix & 0xFF;        // lsb
  bytes_no_fix[1] = (frequency_no_fix >> 8) & 0xFF;    //
  bytes_no_fix[2] = (frequency_no_fix >> 16) & 0xFF;   //
  bytes_no_fix[3] = (frequency_no_fix >> 24) & 0xFF;   // msb

  bytes_fix[0] = frequency_fix & 0xFF;        // lsb
  bytes_fix[1] = (frequency_fix >> 8) & 0xFF;    //
  bytes_fix[2] = (frequency_fix >> 16) & 0xFF;   //
  bytes_fix[3] = (frequency_fix >> 24) & 0xFF;  // msb

  byte message[] = {0xB5, 0x62, 0x06, 0x31, 0x20, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                    bytes_no_fix[0], bytes_no_fix[1], bytes_no_fix[2], bytes_no_fix[3],
                    bytes_fix[0], bytes_fix[1], bytes_fix[2], bytes_fix[3],
                    0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x6F, 0x00, 0x00, 0x00
                   }; //byte array to send to neo gps without checksum bytes ckA and ckB

  //berechnung der checksumme cka und ckb für die ubx nachricht an den gps rx
  unsigned int i;
  byte ckA = 0x00;
  byte ckB = 0x00;              //ckecksummen
  for (i = 2; i < sizeof(message); i = i + 1) { //die ersten beiden elemente des array werden nicht berücksichtigt bei der berechnung der checksumme
    ckA = ckA + message[i];
    ckB = ckB + ckA;
  }

  //Einstellen des timepulse auf beliebige frequenz
  ss.write(message, sizeof(message));   //array: message, Anzahl Bytes zum senden
  ss.write(ckA); //checksumme
  ss.write(ckB);

  delay(1000);
  Serial.println("GPS RX stationärer Betrieb");
  lcd.setCursor(0, 1);         // move cursor to   (0, 1)1. Zeichen, Zeile 1
  lcd.print(" GPS stationaer "); // print message at (0, 1)
  //Einstellen des neo auf staionären Betrieb (erhöht die Genauigkeit)
  //byte array to send to neo gps with checksum bytes ckA and ckB
  //B5 62 06 24 24 00 FF FF 02 03 00 00 00 00 10 27 00 00 05 00 FA 00 FA 00 64 00 2C 01 00 3C 00 00 00 00 00 00 00 00 00 00 00 00 4E 60
  byte message1[] = {0xB5, 0x62, 0x06, 0x24, 0x24, 0x00, 0xFF, 0xFF,
                     0x02, 0x03, 0x00, 0x00, 0x00, 0x00, 0x10, 0x27, 0x00, 0x00, 0x05, 0x00, 0xFA,
                     0x00, 0xFA, 0x00, 0x64, 0x00, 0x2C, 0x01, 0x00, 0x3C, 0x00, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x4E, 0x60
                    };
  ss.write(message1, sizeof(message1));   //array: message, Anzahl Bytes zum senden

  delay(1000);
  Serial.println("GPS RX deaktivieren von SBAS");
  //Deaktivieren von SBAS; wird empfohlen laut ublox für Timinganwendungen
  //byte array to send to neo gps with checksum bytes ckA and ckB
  //B5 62 06 16 08 00 00 01 01 00 00 00 00 00 26 97
  byte message2[] = {0xB5, 0x62, 0x06, 0x16, 0x08, 0x00, 0x00, 0x01,
                     0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x26, 0x97
                    };
  ss.write(message2, sizeof(message2));   //array: message, Anzahl Bytes zum senden
  delay(1000);

  Serial.println("GPS RX speichern der Werte im RAM/ROM....sofern im RX vorhanden");
  //Speichern der Werte im RAM oder ROM vom neo
  //byte array to send to neo gps with checksum bytes ckA and ckB
  //B5 62 06 09 0D 00 00 00 00 00 FF FF 00 00 00 00 00 00 03 1D AB

  byte message3[] = {0xB5, 0x62, 0x06, 0x09, 0x0D, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00,
                     0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x1D, 0xAB
                    };
  ss.write(message3, sizeof(message3));   //array: message, Anzahl Bytes zum senden
  delay(1000);
  lcd.clear();

  //statische Ausgaben an das lcd
  //1. Zeile
  lcd.setCursor(0, 0);         // move cursor to   (0, 0)
  lcd.print(F("OCXO EFC "));    //(F()): Specihern der fixen Daten im Flash-/Programmspeicher; spart Platz im kleineren RAM für die Variaben
  lcd.setCursor(9, 0);         // 10. Zeichen Zeile 1
  lcd.print(F("-.---"));        //
  lcd.setCursor(15, 0);         // 16. Zeichen Zeile 1
  lcd.print(F("V"));
  //2. Zeile
  lcd.setCursor(0, 1); //1. Zeichen, 2. Zeile
  lcd.print(F("Sats:"));
  lcd.setCursor(6, 1); //7. Zeichen, 2. Zeile
  lcd.print(F("--"));
  lcd.setCursor(8, 1); //9. zeichen, 2. Zeile
  lcd.print(F("  "));
  lcd.setCursor(10, 1); //11. Zeichen, 2. Zeile
  lcd.print(F("------"));
  delay(2000);
}

void loop()
{
  smartDelay(800); //read serial data from gps receiver
  getgps(gps) ; // take these GPS parameters and do all calculations
}

//Funktionen
//reads serial data from gps rx
static void smartDelay(unsigned long ms)
{
  unsigned long start = millis();
  do
  {
    while (ss.available())
      gps.encode(ss.read());
  } while (millis() - start < ms);
}

//do calculations with the received data
void getgps(TinyGPSPlus &gps)
{
  //measurement ocxo efc voltage, Pin A0
  float ocxo_spannung = Messe_Spannung();   //oversampling to 15 bit, lsb ==> 0,15mV @ 5V
  ocxo_spannung = ocxo_spannung * 5.00 / 32736.00;
   ocxo_spannung = (float)((int)(ocxo_spannung * 1000)) / 1000.0; //float ocxo_spannung auf drei nachkommastellen begrenzen
  if (ocxo_spannung != ocxo_spannung_old) {
    Serial.print(F("Voltage begrenzt: "));
    Serial.println(ocxo_spannung, 3);
    lcd.setCursor(9, 0);         // 10. Zeichen Zeile 1
    lcd.print(ocxo_spannung, 3);        // print message
    ocxo_spannung_old = ocxo_spannung;
  }

  // Every time anything what we want to know is updated, print this data if there are changes.
  if (gps.satellites.isUpdated() || fix.isUpdated() )
  {
    fix2d3d = *fix.value(); //fix.value ist der zeiger auf die daten mit * kann man auf die daten zugreifen
    sats = gps.satellites.value();

    if (sats != sats_old) {
      Serial.print(F("SATS= ")); Serial.println(sats);
      lcd.setCursor(6, 1); //7. zeichen, 2. Zeile
      if (sats < 10)
      {
        lcd.print(F("0"));
        lcd.print(sats);
      }
      else if (sats >= 10) {
        lcd.print(sats);
      }
      lcd.setCursor(8, 1); //9. zeichen, 2. Zeile
      lcd.print(F("  "));
      sats_old = sats;
    }
    if (fix2d3d == '1' && fix2d3d_old != '1')
    {
      Serial.println(F("no fix"));
      Serial.print(F("Fix_value= ")); Serial.println(fix2d3d);
      Serial.print(F("SATS= ")); Serial.println(sats);
      lcd.setCursor(10, 1); //11. zeichen, 2. Zeile
      lcd.print(F("no fix"));
      lcd.setCursor(6, 1); //7. zeichen, 2. Zeile
      if (sats < 10)
      {
        lcd.print(F("0"));
        lcd.print(sats);
      }
      else if (sats >= 10) {
        lcd.print(sats);
      }
      lcd.setCursor(8, 1); //9. zeichen, 2. Zeile
      lcd.print(F("  "));

      digitalWrite(no_fix_led, HIGH);
      digitalWrite(two_d_fix_led, LOW);
      digitalWrite(three_d_fix_led, LOW);
      fix2d3d_old = fix2d3d;
    }
    else if (fix2d3d == '2' && fix2d3d_old != '2')
    {
      Serial.println(F("2d fix"));
      Serial.print(F("Fix_value= ")); Serial.println(fix2d3d);
      Serial.print(F("SATS= ")); Serial.println(sats);
      lcd.setCursor(10, 1); //11. zeichen, 2. Zeile
      lcd.print(F("2d fix"));
      lcd.setCursor(6, 1); //7. zeichen, 2. Zeile
      if (sats < 10)
      {
        lcd.print(F("0"));
        lcd.print(sats);
      }
      else if (sats >= 10) {
        lcd.print(sats);
      }
      lcd.setCursor(8, 1); //9. zeichen, 2. Zeile
      lcd.print(F("  "));

      digitalWrite(no_fix_led, LOW);
      digitalWrite(two_d_fix_led, HIGH);
      digitalWrite(three_d_fix_led, LOW);
      fix2d3d_old = fix2d3d;
    }
    else if (fix2d3d == '3' && fix2d3d_old != '3')
    {
      Serial.println(F("3d fix"));
      Serial.print(F("Fix_value= ")); Serial.println(fix2d3d);
      Serial.print(F("SATS= ")); Serial.println(sats);
      lcd.setCursor(10, 1); //11. zeichen, 2. Zeile
      lcd.print(F("3d fix"));

      lcd.setCursor(6, 1); //7. zeichen, 2. Zeile
      if (sats < 10)
      {
        lcd.print(F("0"));
        lcd.print(sats);
      }
      else if (sats >= 10) {
        lcd.print(sats);
      }
      lcd.setCursor(8, 1); //9. zeichen, 2. Zeile
      lcd.print(F("  "));

      digitalWrite(no_fix_led, LOW);
      digitalWrite(two_d_fix_led, LOW);
      digitalWrite(three_d_fix_led, HIGH);
      fix2d3d_old = fix2d3d;
    }
    else if (fix2d3d != '1' && fix2d3d != '2' && fix2d3d != '3')
    {
      Serial.println(F("error"));
      Serial.print(F("Fix_value= ")); Serial.println(fix2d3d);
      Serial.print(F(" SATS=")); Serial.println(sats);
      lcd.setCursor(10, 1); //11. zeichen, 2. Zeile
      lcd.print(F("error!"));

      lcd.setCursor(6, 1); //7. zeichen, 2. Zeile
      if (sats < 10)
      {
        lcd.print(F("0"));
        lcd.print(sats);
      }
      else if (sats >= 10) {
        lcd.print(sats);
      }
      lcd.setCursor(8, 1); //9. zeichen, 2. Zeile
      lcd.print("  ");

      digitalWrite(no_fix_led, LOW);
      digitalWrite(two_d_fix_led, LOW);
      digitalWrite(three_d_fix_led, LOW);
      fix2d3d_old = fix2d3d;
    }
  }
}

//ad conversion with oversampling to 15 bit
unsigned long  Messe_Spannung() {
  int j;
  unsigned long analog_val = 0;

  for (j = 1; j <= 1024; j++) {
    analog_val = analog_val + analogRead(A0);    // sensor on analog pin A0; 256 measurements
  }
  analog_val = analog_val >> 5;    // divide by 16 (shift 4 bits right)
  return analog_val;  //
}
