#include <TinyGPS++.h>
#include <SoftwareSerial.h>

// Board: Attiny85;programmed with Arduino UNO
//Attiny85 pins for commuication 
//                        Attiny85
//
//                  +3.3V vom Neo GPS Board
//                  _|_
//                  | |
//                  |R| 15k
//      Elko, 10µF  |_|
//         |  ||     |
// Gnd---- |  ||------------------------Pin1        Pin8  -----+3.3V vom Neo GPS Board
//         |  ||  two_d_fix_led (PB3)---Pin2        Pin7  -----no_fix_led (PB2)
//                three_fix_led (PB4)---Pin3        Pin6  -----TXPin (PB1) geht zum RX Pin vom Neo GPS
//                            GND-------Pin4        Pin5  -----RXPin (PB0) geht zum TXPin vom Neo GPS
//
// Anschlüsse an PB2,3,4: -----330Ohm-----LED-----Gnd
                                 

/*
   Neo M8N gps Empfänger 
   Frequenz am 1pps Ausgang einstellen
   1Hz bis ca. 12 MHz
   Anzeige des Fixes (no fix, 2d fix, 3d fix) über LEDs und serieller Ausgabe sowie der Zahl
   an empfanener Satelliten über serielle Ausgabe
*/
//Globale Variablen und Konstanten

//Frequenz einstellen
byte msb_byte_qrg_no_fix = 0x00;    //hex wert der frequenz wenn gps keinen fix hat : 1Hz dezimal = 0x00 0x00 0x00 0x01 (msb nach lsb) hexadezimal
byte third_byte_qrg_no_fix = 0x00;
byte second_byte_qrg_no_fix = 0x00;
byte lsb_byte_qrg_no_fix = 0x01;

byte msb_byte_qrg_fix = 0x00;     //hex wert der frequenz wenn gps einen fix hat : 100000Hz dezimal = 0x00 0x01 0x86 0xa0 (msb nach lsb) hexadezimal
byte third_byte_qrg_fix = 0x01;
byte second_byte_qrg_fix = 0x86;
byte lsb_byte_qrg_fix = 0xa0;

//Softseriell PINs definieren und Baudrate einstellen
static const int RXPin = PB0, TXPin = PB1;          //pins for software serial
static const uint32_t GPSBaud = 9600;       //baud rate for software serial

//pins für leds
int no_fix_led = PB2, two_d_fix_led = PB3, three_d_fix_led = PB4;    //output pins for the leds

//allg. Variablen
char fix2d3d = '1';



// TinyGPS++ object definieren
TinyGPSPlus gps;

// The serial connection to the GPS device
SoftwareSerial ss(RXPin, TXPin);

//auswertung des fix zustandes: Feld 2 der GPGSA Meldung (custom sequence)
TinyGPSCustom fix(gps, "GPGSA", 2); // $GPGSA sentence, 2th element: char '1' = no fix, char '2' = 2d fix, char '3' = 3d fix


void setup() 
{

  //Serial.begin(SerialBaud);
  ss.begin(GPSBaud);
  
  pinMode(no_fix_led, OUTPUT);          // sets the digital pin as output
  pinMode(two_d_fix_led, OUTPUT);          // sets the digital pin as output
  pinMode(three_d_fix_led, OUTPUT);          // sets the digital pin as output
  digitalWrite(no_fix_led, LOW);
  digitalWrite(two_d_fix_led, LOW);
  digitalWrite(three_d_fix_led, LOW);
 



//array mit den daten zur einstellung der frequenzen am 1pps ausgang ohne abschließende checksumme (2 bytes, ckA und ckB))
byte message[] = {0xB5, 0x62, 0x06, 0x31, 0x20, 0x00, 0x00, 0x01, 0x00, 0x00,0x00, 0x00, 0x00, 0x00,
lsb_byte_qrg_no_fix, second_byte_qrg_no_fix, third_byte_qrg_no_fix, msb_byte_qrg_no_fix,
lsb_byte_qrg_fix, second_byte_qrg_fix, third_byte_qrg_fix, msb_byte_qrg_fix,
0x00,0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x80,0x00, 0x00, 0x00, 0x00, 0x6F, 0x00, 0x00, 0x00}; //byte array to send to neo gps without checksum bytes ckA and ckB

//berechnung der checksumme cka und ckb für die ubx nachricht an den gps rx
int i;
byte ckA = 0x00;
byte ckB = 0x00;              //ckecksummen
for (i = 2; i < sizeof(message); i = i + 1) { //die ersten beiden elemente des array werden nicht berücksichtigt bei der berechnung der checksumme 
  ckA = ckA + message[i];
  ckB = ckB + ckA;
  }

   delay(2000);
//Einstellen des timepulse5 auf beliebige frequenz
  ss.write(message, sizeof(message));   //array: message, Anzahl Bytes zum senden
  ss.write(ckA); //checksumme
  ss.write(ckB);

delay(500);
//Einstellen des neo auf staionären Betrieb (erhöht die Genauigkeit)
  //byte array to send to neo gps with checksum bytes ckA and ckB
  //B5 62 06 24 24 00 FF FF 02 03 00 00 00 00 10 27 00 00 05 00 FA 00 FA 00 64 00 2C 01 00 3C 00 00 00 00 00 00 00 00 00 00 00 00 4E 60
  byte message1[] = {0xB5, 0x62, 0x06, 0x24, 0x24, 0x00, 0xFF, 0xFF,
                     0x02, 0x03, 0x00, 0x00, 0x00, 0x00, 0x10, 0x27, 0x00, 0x00, 0x05, 0x00, 0xFA,
                     0x00, 0xFA, 0x00, 0x64, 0x00, 0x2C, 0x01, 0x00, 0x3C, 0x00, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x4E, 0x60
                    };
  ss.write(message1, sizeof(message1));   //array: message, Anzahl Bytes zum senden
 
  delay(500);
  
  //Deaktivieren von SBAS; wird empfohlen laut ublox für Timinganwendungen
  //byte array to send to neo gps with checksum bytes ckA and ckB
  //B5 62 06 16 08 00 00 01 01 00 00 00 00 00 26 97
  byte message2[] = {0xB5, 0x62, 0x06, 0x16, 0x08, 0x00, 0x00, 0x01,
                     0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x26, 0x97
                    };
  ss.write(message2, sizeof(message2));   //array: message, Anzahl Bytes zum senden
  delay(500);

//Speichern der Werte im RAM oder ROM vom neo
//byte array to send to neo gps with checksum bytes ckA and ckB
//B5 62 06 09 0D 00 00 00 00 00 FF FF 00 00 00 00 00 00 03 1D AB

  byte message3[] = {0xB5, 0x62, 0x06,0x09, 0x0D, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x1D, 0xAB};
  ss.write(message3, sizeof(message3));   //array: message, Anzahl Bytes zum senden
  delay(500);



}

void loop() 
{
  // Every time anything is updated, print everything.
  if (gps.satellites.isUpdated() || fix.isUpdated() )
  {
   fix2d3d = *fix.value(); //fix.value ist der zeiger auf die daten mit * kann man auf die daten zugreifen
   
  if (fix2d3d == '1')
    {
      digitalWrite(no_fix_led, HIGH);
      digitalWrite(two_d_fix_led, LOW);
      digitalWrite(three_d_fix_led, LOW);

    }
  else if (fix2d3d == '2')
    {
       digitalWrite(no_fix_led, LOW);
       digitalWrite(two_d_fix_led, HIGH);
        digitalWrite(three_d_fix_led, LOW);

    }
  else if (fix2d3d == '3')
    {
       digitalWrite(no_fix_led, LOW);
       digitalWrite(two_d_fix_led, LOW);
       digitalWrite(three_d_fix_led, HIGH);

    }
  else
    {
       digitalWrite(no_fix_led, LOW);
       digitalWrite(two_d_fix_led, LOW);
       digitalWrite(three_d_fix_led, LOW);
    }
  }
   
  while (ss.available() > 0)
    gps.encode(ss.read());

}
